import unittest
import numpy as np
from matplotlib import pyplot as plt

from embedded_graph import EmbeddedGraph, NotSingleComponentError, NotPlanarEmbeddingError, EmbeddedSquareGraph, \
    NonSimpleError, SelfLoopError, EdgeNotExistError, EmbeddedHoneycombGraph
from tests.examples import array_A, array_B, array_C, array_D, array_E, array_F, array_G, array_H, array_I, array_J, \
    array_K, array_L


class TestEmbeddedGraph(unittest.TestCase):

    def setUp(self) -> None:
        self.graph_A = EmbeddedGraph(array_A.x, array_A.y, array_A.n1, array_A.n2)
        self.graph_B = EmbeddedGraph(array_B.x, array_B.y, array_B.n1, array_B.n2)
        self.graph_C = EmbeddedGraph(array_C.x, array_C.y, array_C.n1, array_C.n2)
        self.graph_D = EmbeddedGraph(array_D.x, array_D.y, array_D.n1, array_D.n2)
        self.graph_E = EmbeddedGraph(array_E.x, array_E.y, array_E.n1, array_E.n2)
        self.graph_F = EmbeddedGraph(array_F.x, array_F.y, array_F.n1, array_F.n2)
        self.graph_G = EmbeddedGraph(array_G.x, array_G.y, array_G.n1, array_G.n2)
        self.graph_H = EmbeddedGraph(array_H.x, array_H.y, array_H.n1, array_H.n2)
        self.graph_I = EmbeddedGraph(array_I.x, array_I.y, array_I.n1, array_I.n2)
        self.graph_J = EmbeddedGraph(array_J.x, array_J.y, array_J.n1, array_J.n2)
        self.graph_K = EmbeddedGraph(array_K.x, array_K.y, array_K.n1, array_K.n2)
        self.graph_L = EmbeddedGraph(array_L.x, array_L.y, array_L.n1, array_L.n2)
        self.graph_sq = EmbeddedSquareGraph(3, 4, 0.5, 2)

    def test_graph_creation(self):
        with self.assertRaises(NotSingleComponentError):
            EmbeddedGraph([0, 1, 1.5, 1, 0], [0, 0, 0, 1, 1], [0, 4, 1],
                          [1, 0, 2], require_single_component=True)
        with self.assertRaises(NotSingleComponentError):
            EmbeddedGraph([0, 1, 0, 3, 4, 3], [0, 0, 1, 0, 0, 1], [0, 1, 2, 3, 4, 5],
                          [1, 2, 0, 4, 5, 3], require_single_component=True)
        EmbeddedGraph([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 3], [1, 2, 3, 0])
        with self.assertRaises(NotPlanarEmbeddingError):
            EmbeddedGraph([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 3, 0, 1], [1, 2, 3, 0, 2, 3],
                          require_planar_embedding=True)

    def test_get_edge_ids(self):
        self.assertEqual(self.graph_sq.get_edge_ids(0, 1), 0)
        self.assertEqual(self.graph_sq.get_edge_ids(1, 0), 0)
        self.assertTrue(np.all(self.graph_sq.get_edge_ids([2, 4, 0], [5, 3, 3]) == [4, 5, 1]))
        with self.assertRaises(EdgeNotExistError):
            self.graph_sq.get_edge_ids(0, 2)
        with self.assertRaises(EdgeNotExistError):
            self.graph_sq.get_edge_ids(0, 0)
        with self.assertRaises(EdgeNotExistError):
            self.graph_sq.get_edge_ids([2, -1], [5, 0])
        with self.assertRaises(EdgeNotExistError):
            self.graph_sq.get_edge_ids([2, 1, 12], [5, 0, 1])
        self.assertTrue(np.all(self.graph_sq.get_edge_ids([2, 4, 0, 5, 7], [5, 3, 3, 2, 8]) == [4, 5, 1, 4, 12]))
        g = EmbeddedHoneycombGraph(200, 200, 1, 1)
        s = np.random.randint(0, g.edge_count(), 100000)
        n1, n2 = g.node1[s], g.node2[s]
        self.assertTrue(np.all(g.get_edge_ids(n1, n2) == s))

    def test_add_nodes(self):
        g = self.graph_A.add_nodes([1.1, 2.2, 3.3], [4.3, 5.2, 6.1])
        self.assertTrue(np.allclose(g.x, [0, 1, 0, 1, 2, 1, 2, 1.1, 2.2, 3.3]))
        self.assertTrue(np.allclose(g.y, [2, 2, 1, 1, 1, 0, 0, 4.3, 5.2, 6.1]))
        self.assertTrue(g.node_count() == 10)
        g = self.graph_A.add_nodes([], [])
        self.assertTrue(np.allclose(g.x, [0, 1, 0, 1, 2, 1, 2]))
        self.assertTrue(np.allclose(g.y, [2, 2, 1, 1, 1, 0, 0]))
        self.assertTrue(g.node_count() == 7)
        g = self.graph_A.add_nodes(2, 3)
        self.assertTrue(np.allclose(g.x, [0, 1, 0, 1, 2, 1, 2, 2]))
        self.assertTrue(np.allclose(g.y, [2, 2, 1, 1, 1, 0, 0, 3]))
        self.assertTrue(g.node_count() == 8)
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        g.add_nodes([1.1, 2.2, 3.3], [4.3, 5.2, 6.1], in_place=True)
        self.assertTrue(np.allclose(g.x, [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 1.1, 2.2, 3.3]))
        self.assertTrue(g.node_count() == 15)

    def test_add_edges(self):
        g = self.graph_sq.add_edges([0, 1], [4, 5])
        self.assertTrue(np.allclose(g.get_areas(), [0.5, 0.5, 0.5, 0.5, 1, 1, 1, 1, -6]))
        with self.assertRaises(NonSimpleError):
            self.graph_sq.add_edges([0, 1], [4, 4])
        with self.assertRaises(NonSimpleError):
            g = EmbeddedSquareGraph(3, 4, 1, 2)
            g.add_edges([0, 1], [4, 4], in_place=True)
        g = self.graph_sq.add_edges([[0], [1]], [[4], [5]])
        self.assertTrue(np.allclose(g.get_areas(), [0.5, 0.5, 0.5, 0.5, 1, 1, 1, 1, -6]))
        g = self.graph_sq.add_edges([], [])
        self.assertTrue(np.allclose(g.get_areas(), [1, 1, 1, 1, 1, 1, -6]))
        with self.assertRaises(SelfLoopError):
            g = EmbeddedSquareGraph(3, 4, 1, 2)
            g.add_edges([0, 2, 1], [4, 2, 5], in_place=True)
        with self.assertRaises(SelfLoopError):
            self.graph_sq.add_edges([0, 2, 1], [4, 2, 5], in_place=False)
        with self.assertRaises(NotPlanarEmbeddingError):
            g = EmbeddedSquareGraph(3, 4, 1, 2)
            g.add_edges([0, 1, 1], [4, 5, 3], in_place=True, require_planar_embedding=True)
        with self.assertRaises(NotPlanarEmbeddingError):
            self.graph_sq.add_edges([0, 1, 1], [4, 5, 3], in_place=False, require_planar_embedding=True)

    def test_add_nodes_and_edges(self):
        g = self.graph_sq.add_nodes_and_edges([0.25], [1], [0, 1, 3, 4], [12, 12, 12, 12])
        self.assertTrue(np.allclose(g.get_areas(), [0.25, 0.25, 0.25, 0.25, 1, 1, 1, 1, 1, -6]))
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        g.add_nodes_and_edges([0.5], [1], [0, 1, 3, 4], [12, 12, 12, 12], in_place=True)
        self.assertTrue(np.allclose(g.get_areas(), [0.5, 0.5, 0.5, 0.5, 2, 2, 2, 2, 2, -12]))

    def test_remove_nodes(self):
        g = self.graph_sq.remove_nodes([1, 3])
        self.assertTrue(np.allclose(g.get_areas(), [1, 1, 1, -3]))
        self.assertEqual(g.get_num_components(), 2)
        with self.assertRaises(NotSingleComponentError):
            self.graph_sq.remove_nodes([1, 3], require_single_component=True)
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        g.remove_nodes([1, 3], in_place=True)
        self.assertTrue(np.allclose(g.get_areas(), [2, 2, 2, -6]))

    def test_remove_edges(self):
        g = self.graph_sq.remove_edges([0, 4], [1, 3], require_planar_embedding=True, require_single_component=True)
        self.assertTrue(np.allclose(g.get_areas(), [1, 1, 1, 1, -4]))
        g = self.graph_sq.remove_edges(1, 4, require_planar_embedding=True, require_single_component=True)
        self.assertTrue(np.allclose(g.get_areas(), [1, 1, 1,1, 2, -6]))
        g = self.graph_sq.remove_edges([], [], require_planar_embedding=True, require_single_component=True)
        self.assertTrue(np.allclose(g.get_areas(), [1, 1, 1,1, 1, 1, -6]))
        with self.assertRaises(NotSingleComponentError):
            self.graph_sq.remove_edges([0, 4, 3], [1, 3, 0],  require_single_component=True)
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        g.remove_edges([0, 4, 3], [1, 3, 0], in_place=True)
        self.assertTrue(np.all(g.node1 == [1, 1, 2, 3, 4, 4, 5, 6, 6, 7, 7, 8, 9, 10]))
        self.assertTrue(np.all(g.node2 == [2, 4, 5, 6, 5, 7, 8, 7, 9, 8, 10, 11, 10, 11]))

    def test_remove_edges_by_ids(self):
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        g.remove_edges_by_ids([0, 1, 5], in_place=True)
        self.assertTrue(np.all(g.node1 == [1, 1, 2,  3, 4, 4, 5, 6, 6, 7, 7, 8, 9, 10]))
        self.assertTrue(np.all(g.node2 == [2, 4, 5,  6, 5, 7, 8, 7, 9, 8, 10, 11, 10, 11]))

    def test_permute_nodes(self):
        g = EmbeddedSquareGraph(3, 4, 1, 2)
        self.assertTrue(np.all(g.node1 == [0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 6, 6, 7, 7, 8, 9, 10]))
        self.assertTrue(np.all(g.node2 == [1, 3, 2, 4, 5, 4, 6, 5, 7, 8, 7, 9, 8, 10, 11, 10, 11]))
        self.assertTrue(np.all(g._get_faces(include_boundary_faces=False)[0] == [0, 3, 22, 18, 2, 4, 24, 20, 5, 8, 27, 23, 7, 9,
                                                                                 29, 25, 10, 13, 32, 28, 12, 14, 33, 30]))
        self.assertTrue(np.all(g.get_face_nodes(include_boundary_faces=False)[0] == [0, 1, 4, 3, 1, 2, 5, 4, 3, 4, 7, 6, 4, 5, 8, 7, 6, 7, 10, 9, 7, 8, 11, 10]))
        g.permute_nodes([11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0])
        self.assertTrue(np.all(g.node1 == [0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 6, 6, 7, 7, 8, 9, 10]))
        self.assertTrue(np.all(g.node2 == [1, 3, 2, 4, 5, 4, 6, 5, 7, 8, 7, 9, 8, 10, 11, 10, 11]))
        self.assertTrue(np.all(g.get_face_nodes(include_boundary_faces=False)[0] == [11, 10, 7, 8, 10, 9, 6, 7, 8, 7, 4, 5, 7, 6, 3, 4, 5, 4, 1, 2, 4, 3, 0, 1]))
        self.assertTrue(np.all(g._get_faces(include_boundary_faces=False)[0][:8] == [33, 30, 12, 14, 32, 28, 10, 13]))

    def test_permute_faces(self):
        g = EmbeddedSquareGraph(4, 4, 1, 1)
        g.remove_edges([1, 6, 10], [5, 7, 11], in_place=True)
        g.remove_nodes(9, in_place=True)
        g.permute_faces([1, 2, 3, 0])
        self.assertTrue(np.all(g.get_areas() == [3, 4, -9, 2]))
        self.assertTrue(np.all(g.get_face_nodes()[0] == [2, 3, 7, 10, 14, 13, 9, 6, 4, 5, 6, 9, 13, 12, 11, 8, 0, 4,
                                                         8, 11, 12, 13, 14, 10, 7, 3, 2, 1, 0, 1, 2, 6, 5, 4]))
        self.assertTrue(np.all(g.get_face_nodes()[1] == [8, 8, 12, 6]))
        self.assertTrue(np.all(g.cycle_space(include_boundary_faces=False).todense() ==
                               [[0, 0, 0, 1, -1, 1, 0, 0, 0, -1, 1, 0, -1, 1, 0, 0, -1],
                                [0, 0, 0, 0, 0, 0, 1, -1, 1, 1, 0, -1, 1, 0, -1, -1, 0],
                                [1, -1, 1, 0, 1, 0, -1, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0]]))

    def test_map(self):
        mapA = self.graph_A._get_edge_map()
        self.assertTrue(np.all(mapA == [2, 3, 4, 10, 6, 7, 15, 14, 1, 0, 8, 9, 5, 11, 12, 13]))
        mapB = self.graph_B._get_edge_map()
        self.assertTrue(np.all(mapB[array_B.inv] == array_B.inv[[1, 3, 5, 7, 13, 8, 9, 19, 16, 17, 2, 10, 0, 11, 6, 12, 4, 14, 15, 18]]))
        mapC = self.graph_C._get_edge_map()
        self.assertTrue(np.all(
            mapC[array_C.inv] == array_C.inv[[1, 6, 8, 5, 19, 21, 13, 23, 14, 11, 27, 12, 15, 31, 28, 29, 2, 3,
                                              0, 16, 7, 20, 17, 4, 18, 10, 9, 25, 26, 22, 24, 30]]))
        mapD = self.graph_D._get_edge_map()
        self.assertTrue(np.all(mapD == [1, 0]))
        mapE = self.graph_E._get_edge_map()
        self.assertTrue(np.all(mapE[array_E.inv] == array_E.inv[[2, 3, 9, 4, 7, 0, 1, 5, 6, 8]]))
        mapF = self.graph_F._get_edge_map()
        self.assertTrue(np.all(mapF == [3, 5, 11, 7, 8, 10, 1, 2, 0, 4, 6, 9]))
        mapG = self.graph_G._get_edge_map()
        self.assertTrue(np.all(mapG == [6, 10, 21, 8, 9, 12, 13, 15, 16, 19, 17, 2, 0, 1, 11, 3, 4, 5, 14, 7, 18, 20]))
        mapH = self.graph_H._get_edge_map()
        self.assertTrue(np.all(mapH == [2, 8, 7, 5, 11, 10, 1, 0, 6, 4, 3, 9]))
        mapI = self.graph_I._get_edge_map()
        self.assertTrue(np.all(mapI[array_I.inv] == array_I.inv[[4, 29, 3, 25, 26, 11, 9, 35, 31, 10, 32, 15, 16, 41,
                                                                 39, 37, 38, 43, 42, 20, 22, 23, 24, 47, 49, 1, 0, 5,
                                                                 27, 6, 2, 28, 7, 12, 13, 14, 30, 8, 33, 34, 17, 19, 36,
                                                                 40, 18, 21, 44, 45, 46, 48]]))
        mapJ = self.graph_J._get_edge_map()
        self.assertTrue(np.all(mapJ == [4, 8, 17, 18, 11, 12, 13, 9, 19, 16, 1, 3, 0, 2, 6, 10, 5, 14, 7, 15]))
        mapL = self.graph_L._get_edge_map()
        self.assertTrue(np.all(mapL[array_L.inv] == array_L.inv[[3, 5, 10, 9, 13, 11, 12, 2, 7, 0, 4, 1, 8, 6]]))

    def test_get_faces(self):
        face_edges, face_lengths = self.graph_A._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [4, 4, 8]))
        self.assertTrue(np.all(face_edges == [1, 3, 10, 8, 5, 7, 14, 12, 0, 2, 4, 6, 15, 13, 11, 9]))
        face_edges, face_lengths = self.graph_A._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [4, 4]))
        self.assertTrue(np.all(face_edges == [1, 3, 10, 8, 5, 7, 14, 12]))
        face_edges, face_lengths = self.graph_B._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [4, 8, 8]))
        self.assertTrue(np.all(face_edges == [ 6,  9, 17, 15,  0,  2,  3,  7, 19, 18, 14, 11,  1,  4,  8, 16,  5, 13,  12, 10]))
        face_edges, face_lengths = self.graph_B._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [4, 8]))
        self.assertTrue(np.all(face_edges == [6,  9, 17, 15,  1,  4,  8, 16,  5, 13,  12, 10]))
        face_edges, face_lengths = self.graph_C._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [3, 8, 21]))
        self.assertTrue(np.all(face_edges == array_C.inv[[10, 27, 25, 0, 1, 6, 13, 31, 30, 24, 18, 2, 8, 14, 28, 26,
                                                          9, 11, 12, 15, 29, 22, 17, 3, 5, 21, 20, 7, 23, 4, 19, 16]]))
        face_edges, face_lengths = self.graph_C._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [3, 21]))
        self.assertTrue(np.all(face_edges == array_C.inv[[10, 27, 25, 2, 8, 14, 28, 26, 9, 11, 12, 15, 29,
                                                          22, 17, 3, 5, 21, 20, 7, 23, 4, 19, 16]]))
        face_edges, face_lengths = self.graph_D._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [2]))
        self.assertTrue(np.all(face_edges == [0, 1]))
        face_edges, face_lengths = self.graph_D._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == []))
        self.assertTrue(np.all(face_edges == []))
        face_edges, face_lengths = self.graph_E._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [10]))
        self.assertTrue(np.all(face_edges == [0, 3, 4, 7, 6, 1, 2, 9, 8, 5]))
        face_edges, face_lengths = self.graph_E._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == []))
        self.assertTrue(np.all(face_edges == []))
        face_edges, face_lengths = self.graph_F._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [4, 8]))
        self.assertTrue(np.all(face_edges == [1, 5, 10, 6, 0, 3, 7, 2, 11, 9, 4, 8]))
        face_edges, face_lengths = self.graph_F._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [8]))
        self.assertTrue(np.all(face_edges == [0, 3, 7, 2, 11, 9, 4, 8]))
        face_edges, face_lengths = self.graph_G._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [6, 8, 8]))
        self.assertTrue(np.all(face_edges == [2, 21, 20, 18, 14, 11, 0, 6, 13, 1, 10, 17, 5, 12, 3, 8, 16, 4, 9, 19, 7, 15]))
        face_edges, face_lengths = self.graph_G._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [8, 8]))
        self.assertTrue(np.all(face_edges == [0, 6, 13, 1, 10, 17, 5, 12, 3, 8, 16, 4, 9, 19, 7, 15]))
        face_edges, face_lengths = self.graph_H._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [3, 3, 3, 3]))
        self.assertTrue(np.all(face_edges == [0, 2, 7, 1, 8, 6, 3, 5, 10, 4, 11, 9]))
        face_edges, face_lengths = self.graph_H._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [3, 3]))
        self.assertTrue(np.all(face_edges == [0, 2, 7, 3, 5, 10]))
        face_edges, face_lengths = self.graph_I._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [3, 3, 4, 4, 8, 28]))
        self.assertTrue(np.all(face_edges == array_I.inv[[0, 4, 26, 17, 43, 40, 12, 16, 38, 33, 21, 23, 47, 45, 5, 11, 15, 37,
                                              8, 31, 28, 27, 1, 29, 6, 9, 10, 32, 7, 35, 14, 39, 34, 13, 41, 19, 20,
                                              22, 24, 49, 48, 46, 44, 18, 42, 36, 30, 2, 3, 25]]))
        face_edges, face_lengths = self.graph_I._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [3, 3, 4, 4, 8]))
        self.assertTrue(np.all(face_edges == array_I.inv[[0, 4, 26, 17, 43, 40, 12, 16, 38, 33, 21, 23, 47, 45, 5, 11, 15, 37, 8, 31, 28, 27]]))
        face_edges, face_lengths = self.graph_J._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [5, 5, 10]))
        self.assertTrue(np.all(face_edges == [1, 8, 19, 15, 10, 2, 17, 14, 6, 13, 0, 4, 11, 3, 18, 7, 9, 16, 5, 12]))
        face_edges, face_lengths = self.graph_J._get_faces(include_boundary_faces=False)
        self.assertTrue(np.all(face_lengths == [5, 10]))
        self.assertTrue(np.all(face_edges == [2, 17, 14, 6, 13, 0, 4, 11, 3, 18, 7, 9, 16, 5, 12]))
        face_edges, face_lengths = self.graph_K._get_faces(include_boundary_faces=True)
        self.assertTrue(np.all(face_lengths == [3, 3, 3, 3, 4, 6]))
        self.assertTrue(np.all(face_edges == array_K.inv[[0, 3, 13, 1, 6, 16, 4, 18, 14, 5, 19, 15, 7, 8, 10, 20, 2, 9, 21, 17, 12, 11]]))

    def test_get_face_nodes(self):
        face_nodes, face_lengths = self.graph_A.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 2, 3, 1, 3, 5, 6, 4, 0, 1, 3, 4, 6, 5, 3, 2]))
        face_nodes, face_lengths = self.graph_A.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 2, 3, 1, 3, 5, 6, 4]))
        face_nodes, face_lengths = self.graph_B.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [4, 7, 8, 5, 0, 1, 2, 5, 8, 7, 6, 3, 0, 3, 6, 7, 4, 5, 2, 1]))
        face_nodes, face_lengths = self.graph_B.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [4, 7, 8, 5, 0, 3, 6, 7, 4, 5, 2, 1]))
        face_nodes, face_lengths = self.graph_C.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [8, 11, 9, 0, 1, 2, 10, 14, 13, 12, 6, 0, 6, 12, 13, 11, 8, 9, 11, 13, 14, 10, 2, 1, 4, 5, 4, 3, 7, 3, 4, 1]))
        face_nodes, face_lengths = self.graph_C.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [8, 11, 9, 0, 6, 12, 13, 11, 8, 9, 11, 13, 14, 10, 2, 1, 4, 5, 4, 3, 7, 3, 4, 1]))
        face_nodes, face_lengths = self.graph_D.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 1]))
        face_nodes, face_lengths = self.graph_D.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == []))
        face_nodes, face_lengths = self.graph_E.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 3, 4, 5, 2, 1, 2, 5, 4, 3]))
        face_nodes, face_lengths = self.graph_E.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == []))
        face_nodes, face_lengths = self.graph_F.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 2, 3, 1, 0, 1, 2, 0, 3, 2, 1, 3]))
        face_nodes, face_lengths = self.graph_F.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 1, 2, 0, 3, 2, 1, 3]))
        face_nodes, face_lengths = self.graph_G.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 5, 4, 3, 2, 1, 0, 1, 5, 0, 4, 5, 1, 4, 1, 2, 4, 1, 3, 4, 2, 3]))
        face_nodes, face_lengths = self.graph_G.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 1, 5, 0, 4, 5, 1, 4, 1, 2, 4, 1, 3, 4, 2, 3]))
        face_nodes, face_lengths = self.graph_H.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 1, 5, 0, 5, 1, 2, 3, 4, 2, 4, 3]))
        face_nodes, face_lengths = self.graph_H.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 1, 5, 2, 3, 4]))
        face_nodes, face_lengths = self.graph_I.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 3, 4, 11, 15, 12, 7, 12, 13, 8, 16, 18, 19, 17, 1, 6, 11, 12, 7, 8, 3,
                                              2, 0, 4, 3, 8, 9, 10, 5, 10, 9, 14, 9, 8, 13, 12, 16, 17, 19, 20, 19,
                                              18, 16, 12, 15, 11, 6, 1, 2, 3]))
        face_nodes, face_lengths = self.graph_I.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 3, 4, 11, 15, 12, 7, 12, 13, 8, 16, 18, 19, 17, 1, 6, 11, 12, 7, 8, 3, 2]))
        face_nodes, face_lengths = self.graph_J.get_face_nodes(include_boundary_faces=True)
        self.assertTrue(np.all(face_nodes == [0, 2, 4, 3, 1, 0, 3, 2, 1, 4, 0, 1, 2, 0, 4, 2, 3, 4, 1, 3]))
        face_nodes, face_lengths = self.graph_J.get_face_nodes(include_boundary_faces=False)
        self.assertTrue(np.all(face_nodes == [0, 3, 2, 1, 4, 0, 1, 2, 0, 4, 2, 3, 4, 1, 3]))

    def test_get_face_areas(self):
        face_areas = self.graph_A.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [1, 1, -2]))
        face_areas = self.graph_A.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, [1, 1]))
        face_areas = self.graph_B.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [1, -4, 3]))
        face_areas = self.graph_B.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, [1, 3]))
        face_areas = self.graph_C.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [0.5, -16, 15.5]))
        face_areas = self.graph_C.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, [0.5, 15.5]))
        face_areas = self.graph_D.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [0]))
        face_areas = self.graph_D.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, []))
        face_areas = self.graph_E.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [0]))
        face_areas = self.graph_E.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, []))
        face_areas = self.graph_F.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [-1, 1]))
        face_areas = self.graph_G.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [-2, 1, 1]))
        face_areas = self.graph_H.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [0.5, -0.5, 0.5, -0.5]))
        face_areas = self.graph_H.get_areas(include_boundary_faces=False)
        self.assertTrue(np.allclose(face_areas, [0.5, 0.5]))
        face_areas = self.graph_I.get_areas(include_boundary_faces=True)
        self.assertTrue(np.allclose(face_areas, [0.5, 0.5, 1, 1, 3, -6]))


    def test_get_face_centroids(self):
        c_x, c_y = self.graph_A.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [0.5, 1.5, 1]))
        self.assertTrue(np.allclose(c_y, [1.5, 0.5, 1]))
        c_x, c_y = self.graph_A.get_centroids(include_boundary_faces=False)
        self.assertTrue(np.allclose(c_x, [0.5, 1.5]))
        self.assertTrue(np.allclose(c_y, [1.5, 0.5]))
        c_x, c_y = self.graph_B.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [1.5, 1, 5/6]))
        self.assertTrue(np.allclose(c_y, [0.5, 1, 7/6]))
        c_x, c_y = self.graph_B.get_centroids(include_boundary_faces=False)
        self.assertTrue(np.allclose(c_x, [1.5, 5/6]))
        self.assertTrue(np.allclose(c_y, [0.5, 7/6]))
        c_x, c_y = self.graph_C.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [7/3, 2, 185/93]))
        self.assertTrue(np.allclose(c_y, [5/3, 2, 187/93]))
        c_x, c_y = self.graph_C.get_centroids(include_boundary_faces=False)
        self.assertTrue(np.allclose(c_x, [7/3, 185/93]))
        self.assertTrue(np.allclose(c_y, [5/3, 187/93]))
        c_x, c_y = self.graph_F.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [0.5, 0.5]))
        self.assertTrue(np.allclose(c_y, [0.5, 0.5]))
        c_x, c_y = self.graph_G.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [1, 0.5, 1.5]))
        self.assertTrue(np.allclose(c_y, [0.5, 0.5, 0.5]))
        c_x, c_y = self.graph_H.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [1/3, 1/3, 5/3, 5/3]))
        self.assertTrue(np.allclose(c_y, [1/3, 1/3, 2/3, 2/3]))
        c_x, c_y = self.graph_I.get_centroids(include_boundary_faces=True)
        self.assertTrue(np.allclose(c_x, [8/3, 1/3, 3/2, 3/2, 5/6, 7/6]))
        self.assertTrue(np.allclose(c_y, [13/3, 5/3, 5/2, 1/2, 19/6, 31/12]))

    def test_get_num_components(self):
        self.assertEqual(self.graph_A.get_num_components(), 1)
        self.assertEqual(self.graph_B.get_num_components(), 1)
        self.assertEqual(self.graph_C.get_num_components(), 1)
        self.assertEqual(self.graph_D.get_num_components(), 1)
        self.assertEqual(self.graph_E.get_num_components(), 1)
        self.assertEqual(self.graph_F.get_num_components(), 1)
        self.assertEqual(self.graph_G.get_num_components(), 1)
        self.assertEqual(self.graph_H.get_num_components(), 2)
        self.assertEqual(self.graph_I.get_num_components(), 1)
        self.assertEqual(self.graph_J.get_num_components(), 1)
        self.assertEqual(self.graph_K.get_num_components(), 1)

    def test_is_planar_embedding(self):
        self.assertEqual(self.graph_A.is_planar_embedding(), True)
        self.assertEqual(self.graph_B.is_planar_embedding(), True)
        self.assertEqual(self.graph_C.is_planar_embedding(), True)
        self.assertEqual(self.graph_D.is_planar_embedding(), True)
        self.assertEqual(self.graph_E.is_planar_embedding(), True)
        self.assertEqual(self.graph_F.is_planar_embedding(), False)
        self.assertEqual(self.graph_G.is_planar_embedding(), False)
        self.assertEqual(self.graph_H.is_planar_embedding(), False)
        self.assertEqual(self.graph_I.is_planar_embedding(), True)
        self.assertEqual(self.graph_J.is_planar_embedding(), False)
        self.assertEqual(self.graph_K.is_planar_embedding(), True)

    def test_cut_space(self):
        A = self.graph_A.cut_space().todense()
        self.assertTrue(np.all(A == [[-1, -1, 0,  0,  0,  0,  0,  0],
                                     [1,  0, -1,  0,  0,  0,  0,  0],
                                     [0,  1,  0, -1,  0,  0,  0,  0],
                                     [0,  0,  1,  1, -1, -1,  0,  0],
                                     [0,  0,  0,  0,  1,  0, -1,  0],
                                     [0,  0,  0,  0,  0,  1,  0, -1],
                                     [0,  0,  0,  0,  0,  0,  1,  1]]))
        A = self.graph_B.cut_space().todense()
        self.assertTrue(np.all(A[:, array_B.rinv] == [[-1,  0, -1,  0,  0,  0,  0,  0,  0,  0],
                                                      [ 1, -1,  0,  0,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  1,  0, -1,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  0,  1,  0,  0, -1,  0,  0,  0,  0],
                                                      [ 0,  0,  0,  0, -1,  0, -1,  0,  0,  0],
                                                      [ 0,  0,  0,  1,  1,  0,  0, -1,  0,  0],
                                                      [ 0,  0,  0,  0,  0,  1,  0,  0, -1,  0],
                                                      [ 0,  0,  0,  0,  0,  0,  1,  0,  1, -1],
                                                      [ 0,  0,  0,  0,  0,  0,  0,  1,  0,  1]]))

    def test_cycle_space(self):
        A = self.graph_A.cycle_space(include_boundary_faces=False).todense()
        self.assertTrue(np.all(A == [[-1, 1, -1, 1, 0, 0, 0, 0], [0, 0, 0, 0, -1, 1, -1, 1]]))
        A = self.graph_B.cycle_space(include_boundary_faces=False).todense()
        self.assertTrue(np.all(A[:, array_B.rinv] == [[0, 0, 0, 0, -1, 0, 1, -1, 0, 1], [-1, -1, 1, -1, 1, 1, -1, 0, 1, 0]]))
        A = self.graph_C.cycle_space(include_boundary_faces=False).todense()
        self.assertTrue(np.all(A[:, array_C.rinv] == [[0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 1, -1, 0, 0, 0, 0],
                                                      [-1, -1, 1, 0, 0, 0, -1, 0, 1, 1, -1, 1, 0, -1, 1, 1]]))
        A = self.graph_I.cycle_space(include_boundary_faces=True).todense()
        self.assertTrue(np.all(A[:, array_I.rinv] == [[ 1, -1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  1, -1,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  1, -1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  1, -1,  1,  0],
                                                      [ 0,  0, -1, -1,  0,  1, -1,  0,  1,  0,  0,  1, -1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0],
                                                      [-1,  1,  1,  1, -1, -1,  1,  0,  0,  0,  0, -1,  0,  1,  0,  0, -1, -1,  1,  0,  1, -1,  1, -1,  0]]))
        A = self.graph_K.cycle_space(include_boundary_faces=False).todense()
        self.assertTrue(np.all(A[:, array_K.rinv] == [[ 1,  0, -1,  1,  0,  0,  0,  0,  0,  0,  0],
                                                      [ 0,  1,  0,  0,  0, -1,  1,  0,  0,  0,  0],
                                                      [ 0,  0,  0, -1,  1,  0,  0, -1,  0,  0,  0],
                                                      [ 0,  0,  0,  0, -1,  1,  0,  0, -1,  0,  0],
                                                      [ 0,  0,  0,  0,  0,  0,  0,  1,  1, -1,  1]]))
        A = self.graph_L.cycle_space(include_boundary_faces=True).todense()
        self.assertTrue(np.all(A[:, array_L.rinv] == [[ 1,  0, -1,  1,  0,  0,  0],
                                                      [ 0,  1,  0,  0, -1,  1,  0],
                                                      [-1, -1,  1, -1,  1, -1,  0]]))

    def test_adjacency_matrix(self):
        A = self.graph_A.adjacency_matrix().todense()
        self.assertTrue(np.all(A == [[0, 1, 1, 0, 0, 0, 0],
                                     [1, 0, 0, 1, 0, 0, 0],
                                     [1, 0, 0, 1, 0, 0, 0],
                                     [0, 1, 1, 0, 1, 1, 0],
                                     [0, 0, 0, 1, 0, 0, 1],
                                     [0, 0, 0, 1, 0, 0, 1],
                                     [0, 0, 0, 0, 1, 1, 0]]))
        A = self.graph_B.adjacency_matrix().todense()
        self.assertTrue(np.all(A == [[0, 1, 0, 1, 0, 0, 0, 0, 0],
                                     [1, 0, 1, 0, 0, 0, 0, 0, 0],
                                     [0, 1, 0, 0, 0, 1, 0, 0, 0],
                                     [1, 0, 0, 0, 0, 0, 1, 0, 0],
                                     [0, 0, 0, 0, 0, 1, 0, 1, 0],
                                     [0, 0, 1, 0, 1, 0, 0, 0, 1],
                                     [0, 0, 0, 1, 0, 0, 0, 1, 0],
                                     [0, 0, 0, 0, 1, 0, 1, 0, 1],
                                     [0, 0, 0, 0, 0, 1, 0, 1, 0]]))

    def test_get_common_edge(self):
        common_edges = self.graph_A.get_common_edge_of_faces([0, 1, 2], [1, 2, 0])
        self.assertTrue(np.all(common_edges == [-1, 4, 0]))
        common_edges = self.graph_A.get_common_edge_of_faces([[0, 1], [1, 2]], [[1, 0], [2, 1]])
        self.assertTrue(np.all(common_edges == [[-1, -1], [4, 4]]))
        common_edges = self.graph_B.get_common_edge_of_faces([0, 1, 2], [1, 2, 0])
        self.assertTrue(np.all(common_edges == array_B.inv[[7, 0, 4]]))
        common_edges = self.graph_I.get_common_edge_of_faces([0, 4, 1, 2, 3, 4, 5], [1, 1, 2, 3, 4, 5, 0])
        self.assertTrue(np.all(common_edges == [-1, 15, -1, -1, -1, 2, 0]))

        _, orientation = self.graph_A.get_common_edge_of_faces([0, 1, 2], [1, 2, 0], return_orientation=True)
        self.assertTrue(np.all(orientation == [False, False, True]))
        _, orientation = self.graph_A.get_common_edge_of_faces([[0, 1], [1, 2]], [[1, 0], [2, 1]], return_orientation=True)
        self.assertTrue(np.all(orientation == [[False, False], [False, True]]))
        _, orientation = self.graph_B.get_common_edge_of_faces([0, 1, 2], [1, 2, 0], return_orientation=True)
        self.assertTrue(np.all(orientation == [False, True, True]))
        _, orientation = self.graph_I.get_common_edge_of_faces([0, 4, 1, 2, 3, 4, 5], [1, 1, 2, 3, 4, 5, 0], return_orientation=True)
        self.assertTrue(np.all(orientation == [False, True, False, False, False, False, False]))
        common_edges, orientation= self.graph_K.get_common_edge_of_faces([[i] * 6 for i in range(6)], [[0, 1, 2, 3, 4, 5]] * 6, return_orientation=True)
        self.assertTrue(np.all(common_edges == [[-1, -1,  3, -1, -1,  0],
                                                [-1, -1, -1,  5, -1,  2],
                                                [ 3, -1, -1,  4,  7, -1],
                                                [-1,  5,  4, -1,  9, -1],
                                                [-1, -1,  7,  9, -1,  8],
                                                [ 0,  2, -1, -1,  8, -1]]))
        self.assertTrue(np.all(orientation == [[False, False,  True, False, False,  True],
                                               [False, False, False, False, False,  True],
                                               [False, False, False,  True, False, False],
                                               [False,  True, False, False, False, False],
                                               [False, False,  True,  True, False, False],
                                               [False, False, False, False,  True, False]]))

    def test_cycle_space_solve_for_integral_x(self):
        np.random.seed(0)
        def graph_tester(graph):
            s1 = np.random.randint(0, 100, graph.face_count(include_boundary_faces=False))
            s2 = np.random.randint(0, 100, (1000, graph.face_count(include_boundary_faces=False)))
            x = graph._cycle_space_solve_for_integral_x(s1)
            self.assertTrue(np.all(graph.cycle_space(include_boundary_faces=False) @ x.T == s1.T))
            x = graph._cycle_space_solve_for_integral_x(s2)
            self.assertTrue(np.all(graph.cycle_space(include_boundary_faces=False) @ x.T == s2.T))
        graph_tester(self.graph_A)
        graph_tester(self.graph_B)
        graph_tester(self.graph_C)
        graph_tester(self.graph_I)
        graph_tester(self.graph_K)
        graph_tester(self.graph_L)

if __name__ == "__main__":
    unittest.main()


import unittest
import numpy as np
from josephson_circuit import *

from tests.examples import array_B, array_A, array_C, array_D, array_E
from variable_row_array import VarRowArray


class TestArray(unittest.TestCase):

    def setUp(self):
        self.sq_array_h = SquareArray(3, 4, current_direction='x')
        self.sq_array_v = SquareArray(3, 4, current_direction='y')
        self.hc_array_h = HoneycombArray(3, 4, current_direction='x')
        self.hc_array_v = HoneycombArray(3, 4, current_direction='y')
        self.tr_array_h = TriangularArray(3, 4, current_direction='x')
        self.tr_array_v = TriangularArray(3, 4, current_direction='y')
        self.custom1 = Circuit([0, 1, 1.5, 1, 0], [0, 0, 0, 1, 1], [0, 4, 1, 1, 4, 2],
                               [1, 0, 2, 3, 3, 3], external_current_basis=[1, 0, -2, 0, 1])

    def test_array_creation(self):
        with self.assertRaises(NotSingleComponentError):
            Circuit([0, 1, 1.5, 1, 0], [0, 0, 0, 1, 1], [0, 4, 1], [1, 0, 2])
        with self.assertRaises(NotSingleComponentError):
            Circuit([0, 1, 0, 3, 4, 3], [0, 0, 1, 0, 0, 1], [0, 1, 2, 3, 4, 5], [1, 2, 0, 4, 5, 3])
        Circuit([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 3], [1, 2, 3, 0])
        with self.assertRaises(NotPlanarError):
            Circuit([0, 1, 1, 0, 2], [0, 0, 1, 1, 0], [0, 1, 2, 3, 0, 1, 1, 2], [1, 2, 3, 0, 2, 3, 4, 4])
        with self.assertRaises(NotPlanarEmbeddingError):
            Circuit([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 3, 0, 1], [1, 2, 3, 0, 2, 3])
        with self.assertRaises(NoCurrentConservationError):
            Circuit([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 3], [1, 2, 3, 0], external_current_basis=[1, 2, -1, -1])
        Circuit([0, 1], [0, 0], [0], [1], external_current_basis=[1, -1])

    def test_node_count(self):
        self.assertEqual(self.sq_array_h.node_count(), 12)
        self.assertEqual(self.hc_array_h.node_count(), 46)
        self.assertEqual(self.tr_array_h.node_count(), 24)
        self.assertEqual(self.custom1.node_count(), 5)

    def test_junction_count(self):
        self.assertEqual(self.sq_array_h.junction_count(), 3 * 3 + 2 * 4)
        self.assertEqual(self.hc_array_h.junction_count(), 6 * 12 - 4 - 6 - 2)
        self.assertEqual(self.tr_array_h.junction_count(), 6 * 12 - 16 - 6 + 1)
        self.assertEqual(self.custom1.junction_count(), 6)

    def test_face_count(self):
        self.assertEqual(self.sq_array_h.face_count(), 6)
        self.assertEqual(self.hc_array_h.face_count(), (2 * 3 - 1) * (4 - 1))
        self.assertEqual(self.tr_array_h.face_count(), (3 - 1) * (4 * (4 - 0.5)))
        self.assertEqual(self.custom1.face_count(), 2)

    def test_I_base(self):
        self.assertTrue(np.allclose(np.sum(self.sq_array_h.get_external_current_basis()), 0))
        self.assertTrue(np.allclose(np.sum(self.sq_array_v.get_external_current_basis()), 0))
        self.assertTrue(np.allclose(np.sum(self.hc_array_h.get_external_current_basis()), 0))
        self.assertTrue(np.allclose(np.sum(self.hc_array_v.get_external_current_basis()), 0))
        self.assertTrue(np.allclose(np.sum(self.tr_array_h.get_external_current_basis()), 0))
        self.assertTrue(np.allclose(np.sum(self.tr_array_v.get_external_current_basis()), 0))
        def get_in_out(array: Circuit):
            I_base = array.get_external_current_basis()
            return array.x[I_base > 0], array.y[I_base > 0], array.x[I_base < 0], array.y[I_base < 0], np.sum(I_base[I_base > 0])
        self.assertTrue(np.allclose(get_in_out(self.sq_array_h)[0], 0.0))
        self.assertTrue(np.allclose(get_in_out(self.sq_array_h)[2], 2.0))
        self.assertEqual(get_in_out(self.sq_array_h)[4], 4)
        self.assertTrue(np.allclose(get_in_out(self.sq_array_v)[1], 0.0))
        self.assertTrue(np.allclose(get_in_out(self.sq_array_v)[3], 3.0))
        self.assertEqual(get_in_out(self.sq_array_v)[4], 3)

    def test_add_nodes_and_junctions(self):
        with self.assertRaises(NotSingleComponentError):
            self.custom1.add_nodes_and_junctions([2], [2], [], [])
        with self.assertRaises(NodeNotExistError):
            self.custom1.add_nodes_and_junctions([], [], [0], [5])
        custom1 = self.custom1.add_nodes_and_junctions([2], [2], [2, 3], [5, 5])
        self.assertTrue(np.allclose(np.sort(custom1.areas), [0.25, 0.75, 1]))

    def test_remove_nodes(self):
        with self.assertRaises(NoCurrentConservationError):
            sq_array_h = self.sq_array_h.remove_nodes([0], rescale_external_current_basis=False)
        sq_array_h = self.sq_array_h.remove_nodes([0], rescale_external_current_basis=True)
        self.assertEqual(sq_array_h.node_count(), 11)
        self.assertEqual(sq_array_h.junction_count(), 15)
        self.assertEqual(sq_array_h.face_count(), 5)
        self.assertTrue(np.allclose(sq_array_h.get_external_current_basis(),
                                    [0, -1, 4.0/3, 0, -1, 4.0/3, 0, -1, 4.0/3, 0, -1]))
        with self.assertRaises(NoCurrentConservationError):
            sq_array_v = self.sq_array_v.remove_nodes([0], rescale_external_current_basis=False)
        sq_array_v = self.sq_array_v.remove_nodes([0], rescale_external_current_basis=True)
        self.assertEqual(sq_array_v.node_count(), 11)
        self.assertEqual(sq_array_v.junction_count(), 15)
        self.assertEqual(sq_array_v.face_count(), 5)
        self.assertTrue(np.allclose(sq_array_v.get_external_current_basis(),
                                    [1.5, 1.5, 0. , 0. , 0., 0., 0., 0., -1., -1., -1.]))
        sq_array_h = self.sq_array_h.remove_nodes([1,], rescale_external_current_basis=False)
        self.assertEqual(sq_array_h.node_count(), 11)
        self.assertEqual(sq_array_h.junction_count(), 14)
        self.assertEqual(sq_array_h.face_count(), 4)
        sq_array_h = self.sq_array_h.remove_nodes([1, 2, 4, 5], rescale_external_current_basis=True)
        self.assertEqual(sq_array_h.node_count(), 8)
        self.assertEqual(sq_array_h.junction_count(), 9)
        self.assertEqual(sq_array_h.face_count(), 2)
        with self.assertRaises(NotSingleComponentError):
            sq_array_h = self.sq_array_h.remove_nodes([1, 2, 3, 4, 5], rescale_external_current_basis=True)
        sq_array_h = self.sq_array_h.remove_nodes(0, rescale_external_current_basis=True)
        self.assertEqual(sq_array_h.node_count(), 11)
        self.assertEqual(sq_array_h.junction_count(), 15)
        self.assertEqual(sq_array_h.face_count(), 5)
        sq_array_h = self.sq_array_h.remove_nodes([True, False, False, False, False, False, False, False, False, False, False, False],
                                                  rescale_external_current_basis=True)
        self.assertEqual(sq_array_h.node_count(), 11)
        self.assertEqual(sq_array_h.junction_count(), 15)
        self.assertEqual(sq_array_h.face_count(), 5)

    def test_remove_junctions(self):
        sq_array_h = self.sq_array_h.remove_junctions([0])
        self.assertEqual(sq_array_h.node_count(), 12)
        self.assertEqual(sq_array_h.junction_count(), 16)
        self.assertEqual(sq_array_h.face_count(), 5)
        with self.assertRaises(NotSingleComponentError):
            sq_array_h = self.sq_array_h.remove_junctions([0, 8])
        sq_array_v = self.sq_array_v.remove_junctions([0, 3, 5, 15])
        self.assertEqual(sq_array_v.node_count(), 12)
        self.assertEqual(sq_array_v.junction_count(), 13)
        self.assertEqual(sq_array_v.face_count(), 2)
        sq_array_v = self.sq_array_v.remove_junctions([True, False, False, True, False, True, False, False, False,
                                                       False, False, False, False, False, False, True, False])
        self.assertEqual(sq_array_v.node_count(), 12)
        self.assertEqual(sq_array_v.junction_count(), 13)
        self.assertEqual(sq_array_v.face_count(), 2)
        sq_array_v = self.sq_array_v.remove_junctions(1)
        self.assertEqual(sq_array_v.node_count(), 12)
        self.assertEqual(sq_array_v.junction_count(), 16)
        self.assertEqual(sq_array_v.face_count(), 5)

    def test_get_node_coordinates(self):
        self.assertTrue(np.allclose(self.sq_array_h.get_node_coordinates()[0],
                                    [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]))
        self.assertTrue(np.allclose(self.sq_array_h.get_node_coordinates()[1],
                                    [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3]))

    def test_critical_current_factors(self):
        sq_array_h = self.sq_array_h.copy()
        sq_array_h.set_critical_current_factors(2)
        self.assertEqual(sq_array_h.get_critical_current_factors(), 2)
        sq_array_h.set_critical_current_factors([2])
        self.assertEqual(sq_array_h.get_critical_current_factors(), 2)
        sq_array_h.set_critical_current_factors(np.arange(17))
        self.assertTrue(np.allclose(sq_array_h.get_critical_current_factors(), np.arange(17)))

    def test_resistance_factors(self):
        sq_array_h = self.sq_array_h.copy()
        sq_array_h.set_resistance_factors(2)
        self.assertEqual(sq_array_h.get_resistance_factors(), 2)
        sq_array_h.set_resistance_factors([2])
        self.assertEqual(sq_array_h.get_resistance_factors(), 2)
        with self.assertRaises(ValueError):
            sq_array_h.set_resistance_factors(0.0)
        with self.assertRaises(ValueError):
            sq_array_h.set_resistance_factors(-1.0)
        sq_array_h.set_resistance_factors(np.arange(1, 18))
        self.assertTrue(np.allclose(sq_array_h.get_resistance_factors(), np.arange(1, 18)))
        self.assertTrue(np.allclose(sq_array_h.R(), np.arange(1, 18)))

    def test_capacitance_factors(self):
        sq_array_h = self.sq_array_h.copy()
        sq_array_h.set_capacitance_factors(2)
        self.assertEqual(sq_array_h.get_capacitance_factors(), 2)
        sq_array_h.set_capacitance_factors([2])
        self.assertEqual(sq_array_h.get_capacitance_factors(), 2)
        sq_array_h.set_capacitance_factors(0.0)
        self.assertEqual(sq_array_h.get_capacitance_factors(), 0.0)
        with self.assertRaises(ValueError):
            sq_array_h.set_capacitance_factors(-1.0)
        sq_array_h.set_capacitance_factors(np.arange(1, 18))
        self.assertTrue(np.allclose(sq_array_h.get_capacitance_factors(), np.arange(1, 18)))

    def test_C(self):
        sq_array_h = self.sq_array_h.copy()
        sq_array_h.set_capacitance_factors(2)
        sq_array_h.set_beta_C(3)
        self.assertEqual(sq_array_h.get_beta_C(), 3)
        self.assertEqual(sq_array_h.C(), 6)
        sq_array_h.set_capacitance_factors(np.arange(1, 18))
        self.assertTrue(np.allclose(sq_array_h.C(), 3 * np.arange(1, 18)))

    def test_get_faces(self):
        # incomplete
        self.assertEqual(len(self.sq_array_h._get_faces()), self.sq_array_h.face_count())
        self.assertEqual(len(self.tr_array_h._get_faces()), self.tr_array_h.face_count())
        self.assertEqual(len(self.hc_array_v._get_faces()), self.hc_array_v.face_count())
        self.assertEqual(len(self.custom1._get_faces()), self.custom1.face_count())

    def test_face_areas(self):
        pass

    def test_face_centroids(self):
        pass

    def test_face_common_junctions(self):
        pass

    def test_locate_faces(self):
        pass

    def test_approximate_inductance(self):
        pass

    # def test_junction_inductance_matrix(self):
    #     pass
    #
    # def test_beta_C(self):
    #     # get_beta_C
    #     # set_beta_C
    #     # has_capacitance
    #     pass
    #
    # def test_beta_L(self):
    #     # get_beta_L
    #     # set_beta_L
    #     # has_inductance
    #     pass
    #
    # def test_cut_matrix(self):
    #     pass
    #
    # def test_cycle_matrix(self):
    #     pass
    #
    # def test_reduced_cut_matrix(self):
    #     pass
    #
    # def test_permute_nodes(self):
    #     pass
    #
    # def test_permute_nodes_lexsort(self):
    #     pass
    #
    # def test_all_same_junction_self_L(self):
    #     pass
    #
    # def test_junction_hash(self):
    #     pass
    #
    # def test_Ibase(self):
    #     pass
    #
    # def test_is_orthogonal(self):
    #     pass
    #
    # def test_is_single_component(self):
    #     pass
    #
    # def test_circuit_rank(self):
    #     pass
    #
    # def test_IbaseJ(self):
    #     pass
    #
    # def test_Ibase_total(self):
    #     pass
    #
    # def test_compress_np_array_internal(self):
    #     pass
    #
    # def test_assert_junction_id_exists_internal(self):
    #     pass
    #
    # def test_append_quantities_internal(self):
    #     pass
    #
    # def test_prepare_junction_quantity_internal(self):
    #     pass
    #
    # def test_junction_centers_internal(self):
    #     pass
    #
    # def test_junction_lengths_internal(self):
    #     pass
    #
    # def test_junction_inner_internal(self):
    #     pass
    #
    # def test_get_path_junctions_internal(self):
    #     pass
    #
    # def test_reorder_paths_internal(self):
    #     pass
    #
    # def test_assign_path_map_internal(self):
    #     pass
    #

    def test_get_face_areas_and_centroids_internal(self):
        x1, y1 = [0, 1, 1],  [0, 0, 1]
        n1, cnt1 = [0, 1, 2], [3]
        areas1 = Circuit._get_face_areas_2(VarRowArray(cnt1), n1, x1, y1)
        self.assertTrue(np.allclose(areas1, [0.5]))
        cx1, cy1 = Circuit._get_face_centroids_2(VarRowArray(cnt1), n1, x1, y1)
        self.assertTrue(np.allclose(cx1, [2/3]))
        self.assertTrue(np.allclose(cy1, [1/3]))
        x2, y2 = [0, 1, 2, 0, 1],  [0, 0, 0, 1, 1]
        n2, cnt2 = [0, 1, 4, 3, 1, 2, 4], [4, 3]
        areas2 = Circuit._get_face_areas_2(VarRowArray(cnt2), n2, x2, y2)
        self.assertTrue(np.allclose(areas2, [1.0, 0.5]))
        cx2, cy2 = Circuit._get_face_centroids_2(VarRowArray(cnt2), n2, x2, y2)
        self.assertTrue(np.allclose(cx2, [0.5, 4/3]))
        self.assertTrue(np.allclose(cy2, [0.5, 1/3]))
        x3, y3 = [0, 1, 2, 1, 2, 0, 2], [0, 0, 0, 1, 1, 2, 2]
        n3, cnt3 = [0, 1, 3, 4, 6, 5, 1, 2, 4, 3, 1, 3, 4, 2], [6, 4, 4]
        areas3 = Circuit._get_face_areas_2(VarRowArray(cnt3), n3, x3, y3)
        self.assertTrue(np.allclose(areas3, [3.0, 1.0, -1.0]))
        cx3, cy3 = Circuit._get_face_centroids_2(VarRowArray(cnt3), n3, x3, y3)
        self.assertTrue(np.allclose(cx3, [5/6, 1.5, 1.5]))
        self.assertTrue(np.allclose(cy3, [7/6, 0.5, 0.5]))

    def test_make_cycle_structure_internal(self):
        mapA = Circuit._make_cycle_structure2(array_A.x, array_A.y, array_A.n1, array_A.n2)
        self.assertTrue(np.all(mapA == [2, 3, 4, 10, 6, 7, 15, 14, 1, 0, 8, 9, 5, 11, 12, 13]))
        mapB = Circuit._make_cycle_structure2(array_B.x, array_B.y, array_B.n1, array_B.n2)
        self.assertTrue(np.all(mapB == [1, 3, 5, 7, 13, 8, 9, 19, 16, 17, 2, 10, 0, 11, 6, 12, 4, 14, 15, 18]))
        mapC = Circuit._make_cycle_structure2(array_C.x, array_C.y, array_C.n1, array_C.n2)
        self.assertTrue(np.all(mapC == [1, 6, 8, 5, 19, 21, 13, 23, 14, 11, 27, 12, 15, 31, 28, 29, 2, 3, 0, 16, 7, 20, 17, 4, 18, 10, 9, 25, 26, 22, 24, 30]))
        mapD = Circuit._make_cycle_structure2(array_D.x, array_D.y, array_D.n1, array_D.n2)
        self.assertTrue(np.all(mapD == [1, 0]))
        mapE = Circuit._make_cycle_structure2(array_E.x, array_E.y, array_E.n1, array_E.n2)
        self.assertTrue(np.all(mapE == [2, 3, 9, 4, 7, 0, 1, 5, 6, 8]))

    def test_retrace_paths_internal(self):
        paths1 = np.array([[1, 4, 6, 8, 8, 4, 3, 5, 7, 3]])
        retraced_paths1, retraced_path_lengths1 = Circuit._retrace_paths(paths1)
        self.assertTrue(np.all(retraced_paths1 == [1, 4, 3]))
        self.assertTrue(np.all(retraced_path_lengths1 == [3]))
        paths2 = np.array([[1, 4, 6, 8, 8, 4, 3, 5, 7, 3], [2, 3, 9, 5, 4, 3, 7, 10, 7, 8]])
        retraced_paths2, retraced_path_lengths2 = Circuit._retrace_paths(paths2)
        self.assertTrue(np.all(retraced_paths2 == [1, 4, 3, 2, 3, 7, 8]))
        self.assertTrue(np.all(retraced_path_lengths2 == [3, 4]))
        paths3 = np.array([[4, 6, 4, 5]])
        retraced_paths3, retraced_path_lengths3 = Circuit._retrace_paths(paths3)
        self.assertTrue(np.all(retraced_paths3 == [4, 5]))
        self.assertTrue(np.all(retraced_path_lengths3 == [2]))

    def test_store_terminated_paths_internal(self):
        p, pl = [], []
        pl, p = Circuit._store_terminated_paths([[2, 4, 5], [7, 6, 5], [4, 5, 2]], pl, p)
        self.assertTrue(np.all(pl == [3, 3]))
        self.assertTrue(np.all(p == [2, 4, 5, 5, 7, 6]))
        pl, p = Circuit._store_terminated_paths([[6, 5, 7, 8], [1, 3, 2, 4], [5, 4, 6, 4], [7, 8, 6, 5]], pl, p)
        self.assertTrue(np.all(pl == [3, 3, 4, 4, 2]))
        self.assertTrue(np.all(p == [2, 4, 5, 5, 7, 6, 1, 3, 2, 4, 5, 7, 8, 6, 4, 5]))
        pl, p = Circuit._store_terminated_paths([[6, 5, 7, 8, 1, 3, 2, 4], [5, 4, 6, 4, 7, 8, 6, 5]], pl, p)
        self.assertTrue(np.all(pl == [3, 3, 4, 4, 2]))
        self.assertTrue(np.all(p == [2, 4, 5, 5, 7, 6, 1, 3, 2, 4, 5, 7, 8, 6, 4, 5, 1, 3, 2, 4, 6, 5, 7, 8]))

    def test_construct_faces_internal(self):
        pass


    #
    # def check_planarity_internal(self):
    #     pass
    #
    # def test_junction_remove_mask_internal(self):
    #     pass
    #
    # def test_apply_matrix_format_internal(self):
    #     pass
    #
    # def test_is_sparse_diag_internal(self):
    #     pass
    #
    # def test_is_diagonal_internal(self):
    #     pass
    #
    # def test_is_sparse_symmetric_internal(self):
    #     pass
    #
    # def test_is_symmetric_internal(self):
    #     pass


if __name__ == "__main__":
    unittest.main()



import time

import scipy.sparse
from scipy.spatial import KDTree
import numpy as np
import matplotlib.pyplot as plt
import matplotlib

from pyjjasim.static_problem import lobpcg_test_negative_definite

matplotlib.use("TkAgg")
from pyjjasim import *

N = 30


# a = EmbeddedSquareGraph(N, N)
# # print(time.perf_counter() - tic)
# # tic = time.perf_counter()
# # b, map = a.extended_dual_graph()
# # print(time.perf_counter() - tic)
# # print(map)
# tic = time.perf_counter()
# fig, ax = a.plot(figsize=[20, 10], show_face_ids=True, cycles="l_cycles")
# print(time.perf_counter() - tic)
# # b.plot(ax=ax, show_face_ids=True)
#
#
#
#
# plt.show()

# N = a.junction_count()
# L1 = scipy.sparse.rand(N, N, 0.5)
# L1 = L1 + L1.T + 5 * scipy.sparse.eye(N, N)
# np.set_printoptions(linewidth=1000000)
# a.set_inductance_factors(L1)
# L2 = scipy.sparse.hstack([ scipy.sparse.rand(5, N, 0.5), 3 * scipy.sparse.eye(5)])
# L3 = scipy.sparse.rand(5, 5, 0.5)
# L3 = L3 + L3.T + 5 * scipy.sparse.eye(5)
# print(L2.shape)
# b = a.add_nodes_and_junctions([3, 3, 3], [0, 1, 2],
#                               [2, 5, 8, 9, 10],
#                               [9, 10,11,10, 11],
#                               inductance_factors=L3)
# print(np.round(b.inductance_factors.todense(), 3))
# a.plot()
# plt.show()
# N = 5
# A = scipy.sparse.rand(N, N, 1)
# A = A.tocsc()
# print(A[[True, False, False, True, True], :][:, [False, False, False, True, True]])
# print(a.graph.get_areas())


# def flipp(n1, n2):
#     flip = n1 > n2
#     nn1, nn2 =  np.where(flip, n2, n1), np.where(flip, n1, n2)
#     s = np.lexsort((nn2, nn1))
#     print("s",  s)
#     return  nn1[s], nn2[s], flip, s
#
# def flipi(N1, N2, flip, s):
#     inv_s = np.argsort(s)
#     print("inv s",  inv_s)
#     nn1, nn2 = N1[inv_s], N2[inv_s]
#     n1, n2 = np.where(flip, nn2, nn1), np.where(flip, nn1, nn2)
#     return  n1, n2
#
#
#
# n1 = np.array([2, 4, 6, 0, 1, 2])
# n2 = np.array([1, 3, 0, 4, 5, 3])
#
# print(n1)
# print(n2)
#
# N1, N2, flip, s = flipp(n1, n2)
#
# print(N1)
# print(N2)
#
#
# n1r, n2r = flipi(N1, N2, flip, s)
#
# print(n1r)
# print(n2r)

# fig, ax = plt.subplots()
# out = ax.plot([1, 2, 3], [1, 4, 9])
# o = out[0]
# o.set_xdata([3, 4, 5])
# print([method_name for method_name in dir(o)
#                   if callable(getattr(o, method_name))])
# plt.show()

# def nearest_anchor(x):
#     b = np.log(x) / np.log(10)
#     decade = np.floor(b)
#     B = b - decade
#     s = np.array([0, np.log(2) / np.log(10), np.log(5) / np.log(10), 1])
#     sub = np.argmin(np.abs(B - s))
#     if sub == 0:
#         return 10 ** decade
#     if sub == 1:
#         return 2 * 10 ** decade
#     if sub == 2:
#         return 5 * 10 ** decade
#     if sub == 3:
#         return 10 ** (decade + 1)
#
# print(nearest_anchor(14.8))
#
# a = np.zeros(3)
# print(np.atleast_2d(a).shape)

# ar = SquareArray(5, 5)
# M = ar.get_cycle_matrix()
# A = M @ M.T
# out = lobpcg_test_negative_definite(-A, preconditioner=None, accept_ratio=10, maxiter=200)
# print(out)
# print(len(out[1]))
# print(len(out[2]))


"""
EXAMPLE 4: Parameter optimization

Compute maximal frustration bounds and maximal current 
"""

if __name__ == "__main__":
    # define arrays
    array = SquareArray(12, 12)
    array.set_inductance_factors(1)
    Ih, Iv = 1.0 * array.horizontal_junctions(), 1.0 * array.vertical_junctions()

    # compute frustration bounds for zero vortex state
    prob = StaticProblem(array, current_sources=0)
    (smallest_f, largest_f), (s_config, l_config), (i1, i2) = prob.compute_frustration_bounds()
    s_config.plot(title=f"minimal frustration in zero vortex state (f={np.round(smallest_f, 4)})")
    print(i2)
    scheme=1
    # prec = stability_get_preconditioner(array, prob.get_current_phase_relation(), scheme=scheme)
    st = {"scheme": scheme, "algorithm": 0, "preconditioner": "auto"}
    prob = StaticProblem(array, current_sources=0)
    (smallest_f, largest_f), (s_config, l_config), (i1, i2) = prob.compute_frustration_bounds(stability_parameters=st)
    s_config.plot(title=f"minimal frustration in zero vortex state (f={np.round(smallest_f, 4)})")
    print(i2)


    # # compute maximal current
    # prob = StaticProblem(array, frustration=0, current_sources=Iv)
    # I_factor, net_I, max_I_config, info = prob.compute_maximal_current()
    # np.set_printoptions(linewidth=10000000)
    # print(f"largest current factor {I_factor} (corresponding to net current of  {net_I}) at which the zero-vortex state exists at zero frustration")
    # print(info)
    # max_I_config.plot(title=f"maximal current in zero vortex state (net_I={np.round(net_I, 4)})")
    #
    #
    # # compute maximal current
    # n = np.zeros(array.face_count())
    # n[array.locate_faces(5.5, 5.5)] = 1
    # prob = StaticProblem(array, frustration=0.05, current_sources=Iv, vortex_configuration=n)
    # I_factor, net_I, max_I_config, info = prob.compute_maximal_current(require_vortex_configuration_equals_target=False)
    # np.set_printoptions(linewidth=10000000)
    # print(f"largest current factor {I_factor} (corresponding to net current of  {net_I}) at which the zero-vortex state exists at zero frustration")
    # print(info)
    # max_I_config.plot(title=f"maximal current in zero vortex state (net_I={np.round(net_I, 4)})")

    #
    # # compute extermum in Is-f space using compute_stable_region()
    # prob = StaticProblem(array, frustration=0, current_sources=Iv)
    # f, net_I, _, _ = prob.compute_stable_region()
    # plt.subplots()
    # plt.plot(f, net_I)
    # plt.xlabel("frustration")
    # plt.ylabel("net I")
    # plt.title("Region in f-I parameter space where the zero-vortex state is stable")
    #
    # # compute direction dependent maximal current using compute_maximal_parameter()
    # prob = StaticProblem(array)
    # angles = np.linspace(0, 2*np.pi, 33)
    # I_factor = np.zeros(len(angles))
    # for i, angle in enumerate(angles):
    #     prob_func = lambda x: prob.new_problem(current_sources=x * (np.cos(angle) * Ih + np.sin(angle) * Iv))
    #     I_factor[i], _, _, _ = compute_maximal_parameter(prob_func)
    # plt.subplots()
    # plt.plot(np.cos(angles) * I_factor, np.sin(angles) * I_factor, marker="o")
    # plt.xlabel("horizontal current")
    # plt.ylabel("vertical current")
    # plt.title("angle dependent maximal current for zero-vortex state")
    # plt.show()



# import scipy.sparse.linalg
#
# from pyjjasim import *
# ar = HoneycombArray(4, 4)
#
# A = ar.get_cycle_matrix()
# A = A @ A.T
#
# np.set_printoptions(linewidth=100000)
#
# f = scipy.sparse.linalg.splu(A, diag_pivot_thresh=0)
# Up = (f.L @ scipy.sparse.diags(f.U.diagonal())).T
# print(np.allclose((Up - f.U).data, 0))

import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt

from pyjjasim import *

if __name__ == "__main__":

    square_array = SquareArray(3, 4)
    honeycomb_array = HoneycombArray(3, 4)
    np.set_printoptions(linewidth=100000, threshold=100000000)

    plt.show()

    triangular_array = TriangularArray(4, 3)

    print(square_array)
    print(square_array.current_base(np.pi/4))

    print(honeycomb_array)
    print(honeycomb_array.current_base(0, type="junction"))
    print(np.sign(np.array([False, True]).astype(int)))
    # honeycomb_array.current_base(0, type="node")
    CircuitPlot(triangular_array, arrow_data=triangular_array.current_base(np.pi/4, type="junction")).make()
    plt.show()
    # frustration factor
    f = 0.01

    # define physical problems
    # # prob_sq = StaticProblem(square_array, frustration=f, current_sources=square_array.horizontal_junctions())
    # # prob_hc = StaticProblem(honeycomb_array, frustration=f, current_sources=honeycomb_array.horizontal_junctions())
    # # prob_tr = StaticProblem(triangular_array, frustration=f, current_sources=triangular_array.horizontal_junctions())
    #
    # # square_array.plot()
    # # honeycomb_array.plot()
    # # triangular_array.plot()
    # # plt.show()
    # #
    # # # compute maximal current
    # # _, _, config_sq, _ = prob_sq.compute_maximal_current()
    # # _, _, config_hc, _ = prob_hc.compute_maximal_current()
    # # _, _, config_tr, _ = prob_tr.compute_maximal_current()
    # #
    # # # plot result
    # # config_sq.plot()
    # # config_hc.plot()
    # # config_tr.plot()
    # # plt.show()


def dual_graph(self, bounding_box=None):
    def get_e_edge(A, idx_nb, idx_b):
        mat = A[idx_nb, :].multiply(A[idx_b, :]).tocoo()
        _, index = np.unique(mat.row, return_index=True)
        return mat.col[index]

    def get_new_coordinates(n1, n2, x, y, xc, yc, bounding_box):
        L = len(xc)
        e = np.ones(L)
        X = 0.5 * (x[n1] + x[n2])
        Y = 0.5 * (y[n1] + y[n2])
        bx1, bx2 = bounding_box[0], bounding_box[0] + bounding_box[2]
        by1, by2 = bounding_box[1], bounding_box[1] + bounding_box[3]
        l = np.stack([(bx1 - xc) / (X - xc), (bx2 - xc) / (X - xc),
                      (by1 - yc) / (Y - yc), (by2 - yc) / (Y - yc)], axis=0)
        Xo = np.stack([bx1 * e, bx2 * e, xc + l[2, :] * (X - xc), xc + l[3, :] * (X - xc)], axis=0)
        Yo = np.stack([yc + l[0, :] * (Y - yc), yc + l[1, :] * (Y - yc), by1 * e, by2 * e], axis=0)
        idx = np.argmax(1 / np.array(l), axis=0)
        return Xo[(idx, np.arange(L))], Yo[(idx, np.arange(L))]

    x, y = self.coo()
    X, Y = self.get_l_cycle_centroids()
    n1, n2 = self.node1, self.node2
    if self.get_num_components() != 1:
        raise ValueError("number of components must be 1")

    k = 0.01
    if bounding_box is None:
        xl, xh, yl, yh = np.min(X), np.max(X), np.min(Y), np.max(Y)
        bounding_box = [(1 + k) * xl - k * xh, (1 + k) * yl - k * yh, (1 + 2 * k) * (xh - xl), (1 + 2 * k) * (yh - yl)]

    Al = self.l_cycle_matrix(_permute=False)

    adjl = (Al @ Al.T).tocoo()
    mask = adjl.row < adjl.col

    N1, N2 = adjl.row[mask], adjl.col[mask]  # in range(Nfl)

    b_mask = ~self._non_boundary_mask()

    edge_cases = b_mask[N1] | b_mask[N2]
    b_is_N1 = b_mask[N1][edge_cases]
    N1e = N1[edge_cases]
    N2e = N2[edge_cases]
    Ne_nb = np.where(b_is_N1, N2e, N1e)
    Ne_b = np.where(b_is_N1, N1e, N2e)
    e_edge = get_e_edge(Al, Ne_nb, Ne_b)
    xn, yn = get_new_coordinates(n1[e_edge], n2[e_edge], x, y, X[Ne_nb], Y[Ne_nb], bounding_box)

    new_node_idx = len(X) + np.arange(len(xn))

    N1e[b_is_N1] = new_node_idx[b_is_N1]
    N2e[~b_is_N1] = new_node_idx[~b_is_N1]
    N1[edge_cases] = N1e
    N2[edge_cases] = N2e

    X = np.append(X, xn)
    Y = np.append(Y, yn)

    return EmbeddedGraph(X, Y, N1, N2)

    def extended_graph(self, bounding_box=None, box_extend_factor=1.2):
        self._assign_faces()
        self._assign_boundary_faces()
        x, y = self.coo()
        n1, n2 = self.node1, self.node2
        if self.get_num_components() != 1:
            raise ValueError("number of components must be 1")
        k = box_extend_factor - 1
        if bounding_box is None:
            xl, xh, yl, yh = np.min(x), np.max(x), np.min(y), np.max(y)
            bounding_box = [(1+k) * xl - k * xh, (1+k) * yl - k * yh, (1+2*k) * (xh - xl), (1+2*k) *(yh - yl)]
        nodes = np.unique(self.faces_v_array.delete_rows(self.face_nodes, np.flatnonzero(self._non_boundary_mask())))
        X, Y = x[nodes], y[nodes]
        bx1, bx2 = bounding_box[0], bounding_box[0] + bounding_box[2]
        by1, by2 = bounding_box[1], bounding_box[1] + bounding_box[3]
        L = len(nodes)
        e = np.ones(L)
        dist = np.stack([np.abs(X - bx1), np.abs(X - bx2), np.abs(Y - by1), np.abs(Y - by2)], axis=0)
        Xo = np.stack([bx1 * e, bx2 * e, X, X], axis=0)
        Yo = np.stack([Y, Y, by1 * e, by2 * e], axis=0)
        idx = np.argmin(dist, axis=0)
        Xo, Yo = Xo[(idx, np.arange(L))], Yo[(idx, np.arange(L))]
        ido = np.arange(len(Xo)) + len(x)
        X = np.concatenate((x, Xo, [bx1, bx2, bx2, bx1]))
        Y = np.concatenate((y, Yo, [by1, by1, by2, by2]))
        P = len(x) + len(ido)

        sxl = ido[idx == 2][np.argsort(Xo[idx == 2])]
        sxh = ido[idx == 3][np.argsort(Xo[idx == 3])]
        syl = ido[idx == 0][np.argsort(Yo[idx == 0])]
        syh = ido[idx == 1][np.argsort(Yo[idx == 1])]

        N1 = np.concatenate((n1, nodes, np.append(sxl, [P+ 1]), np.append(sxh, [P+2]), np.append(syl, [P+3]), np.append(syh, [P+2])))
        N2 = np.concatenate((n2, ido,   np.append([P], sxl),    np.append([P+3], sxh), np.append([P], syl),   np.append([P+1], syh)))
        return EmbeddedGraph(X, Y, N1, N2)

    def extended_dual_graph(self, bounding_box=None, box_extend_factor=1.2):
        dual_graph = self.extended_graph(bounding_box=bounding_box,
                                         box_extend_factor=box_extend_factor).face_dual_graph()
        mapping = dual_graph.locate_faces(*self.coo())
        return dual_graph, mapping
